-- main.lua (of bug_tv app)

sversion = '20200325.1'

-- Author: Stan Sieler, sieler@allegro.com

--[[
---------------------------------------------------------------
This app demonstrates two bugs in Corona SDK's handling
of TableViews and isVisible.
(Duplicated on simulator (2020.3575, and Android)

We create a tableview, and then a button in the same screen area.  
The button is set invisible.  
If you tap on a row of the table, we make the table invisible 
and make the button ('Tap Me') visible.

NOW ...if you tap on 'Tap Me', you will usually see a spurious 
event passed to the tableview touch handler (bug 1)!  
(See about line 409, in middle of 'handle_touch_row')
Note that handle_touch_row has code to notice and 
workaround the spurious touch (using time-of-day,
which is a kludge).

The spurious touch will be noted on the Corona console,
and in the app's text message area.  

The spurious touch will report the row that was "under" 
the 'Tap Me' button.  Since the button slightly overlaps
two rows, you'll see a different row being tapped depending
upon whether you tap the top or bottom half of the button.
The row number also depend upon whether or not you scrolled
the table before tapping.

Sometimes, the tableview doesn't even become visible (bug 2).
And, inf cat, sometimes, you can tap dozens of times before 
the tableview becomes visible!

The app has in-app help to assist you in testing.

Note: this app does not ever test the value of isVisible,
because there is a known bug (reported 2 years ago) in
that (i.e., sometimes an incorrect value is returned.)
---------------------------------------------------------------
]]

-- "tv" stands for "TableView"
-- "bh" for "Button Handler"
-- "t_" for "Text"

widget = require ('widget')

------ globals for display data 

contH              = display.contentHeight
contH2             = math.floor (display.contentHeight / 2)
contW              = display.contentWidth
contW2             = math.floor (display.contentWidth / 2)

--- user globals

avoid_bounce_time  = 0
b_help = nil
b_reset = nil
b_tap_me = nil
t_help = nil
t_msg = nil
t_numbers_bottom = nil
t_numbers_side = nil
t_title = nil
tv_list = nil

-----------------------------------------------------------------
function bh_button_tapped (event)

   if event.phase == "ended" then
 
      print ("Button tapped")

      show_msg ("Button tapped")

      b_tap_me.isVisible = false
      tv_list.isVisible = true

         -- without the following, tapping on the 'Tap Me' 
         -- button will result in handle_touch_row being
         -- spuriously called (the console display will
         -- show "handle_touch_row 4" (or 5, depending upon
         -- whether you tapped on the top edge of the button
         -- or the bottom edge!)
         -- When the tableview is visible, note that the
         -- "Tap Me" button will be overlaid on top of
         -- rows 4 and 5 (hence the 4 or 5 mentioned above!)

         -- the following, of course, doesn't *prevent*
         -- the spurious call.  It simply allows handle_touch_row
         -- to say "my gosh, this call is probably spurious...
         -- I'll ignore it".

      avoid_bounce_time = system.getTimer ()
   end

   return true

end -- bh_button_tapped proc
-----------------------------------------------------------------
function bh_help_tapped (event)    -- show help text

   if event.phase == "ended" then

      set_all_invisible ()

      t_help.isVisible = true 

      t_help.text = [[
This app demonstrates one or two
bugs in Corona SDK with respect to
'isVisible' and event handling.

We create a tableview, and then a button
in the same screen area.  The button is
invisible.  If you tap on a row of the table,
we make the table invisible and the button
visible.
NOW ...if you tap on 'Tap Me', you will
usually see a spurious event passed to the
tableview touch handler (bug 1)!  This will 
be noted on the console, and in the text
message area.  Sometimes, the tableview 
doesn't even become visible (bug 2).
(Sometimes, you can tap dozens of times
before the tableview becomes visible!)

Tapping 'Reset' restores visibility of 
everything to the starting state.

Note: if you tap on a message on the
screen, it will be erased for you.]]

   end

   return true

end -- bh_help_tapped proc
-----------------------------------------------------------------
function bh_help_text (event)    -- when 'help' text is tapped, it's erased & set invisible

   if event.phase == "ended" then
      t_help.isVisible = false
      t_help.text = ""
      
      reset_stuff ()   -- set visibility back to starting state
   end

   return true

end -- bh_help_text proc
-----------------------------------------------------------------
function bh_msg (event)    -- when 'msg' text is tapped, it gets erased

   if event.phase == "ended" then
 
      t_msg.text = ""
   end

   return true

end -- bh_msg proc
-----------------------------------------------------------------
function bh_reset_tapped (event)

   if event.phase == "ended" then
 
      print ("reset tapped")

      show_msg ([['Reset' tapped:
table view set to Visible,
'Tap Me' set to inVisible]])

      reset_stuff ()

   end

   return true

end -- bh_reset_tapped proc
-----------------------------------------------------------------
function bh_title_text (event)    -- when 'title' text is tapped, it's erased

   if event.phase == "ended" then
      t_title.text = ""
   end

   return true

end -- bh_title_text proc
-----------------------------------------------------------------
function create_help ()

    local x, y

   x = contW2
   y = contH2

   t_help = display.newText ( "help text", x, y, "Helvetica", 30)
   
   t_help:addEventListener ("touch", bh_help_text)
   t_help.isVisible = false

   x = contW - 100
   y = contH - 15

   b_help = widget.newButton ( {
         defaultFile = "b120_base.png",
         overFile = "b120_over.png",
         label = "Help",
         emboss = true,
         fontSize = 30,
         onRelease = bh_help_tapped,
         x = x,
         y = y
         } )

end -- create_help proc
-----------------------------------------------------------------
function create_msg ()

    local x, y

   x = contW2
   y = contH - 180
   t_msg = display.newText (
   [[Tap any row.  A 'Tap Me' button
   will appear.  Tap it.  Then, look 
   here after that tap.]],
                            x, y, "Helvetica", 30)

   t_msg:addEventListener ("touch", bh_msg)

end -- create_msg proc
-----------------------------------------------------------------
function create_numbers_on_bottom ()

      -- creates list of numbers (labelling pixel positions) 
      -- across the bottom of the screen.
      -- The numbers are a visual aid in debugging screen
      -- layout during development.
      -- NOTE: the upper left of the number is at the pixel labelled

   if t_numbers_bottom == nil then
      t_numbers_bottom = {}
      for ktr = 0, contW, 50 do
         t_numbers_bottom [ktr] = display.newText (tostring (ktr), ktr, contH - 50, "Helvetica", 15)
         t_numbers_bottom [ktr].anchorX = 0
      end
   end

end -- create_numbers_on_side proc
-----------------------------------------------------------------
function create_numbers_on_side ()

      -- creates list of numbers (labelling pixel positions) 
      -- down the right side of the screen.
      -- The numbers are a visual aid in debugging screen
      -- layout during development.
      -- NOTE: the upper left of the number is at the pixel labelled

   if t_numbers_side == nil then
      t_numbers_side = {}
      for ktr = 100, contH, 50 do
         t_numbers_side [ktr] = display.newText (tostring (ktr), contW - 50, ktr)
         t_numbers_side [ktr].anchorX = 0
         t_numbers_side [ktr].anchorY = 0
      end
   end

end -- create_numbers_on_side proc
-----------------------------------------------------------------
function create_reset ()  -- creates the 'reset' button (lower left)

   local x, y

   x = 100
   y = contH - 15

   b_reset = widget.newButton ( {
         defaultFile = "b120_base.png",
         overFile = "b120_over.png",
         label = "Reset",
         emboss = true,
         fontSize = 30,
         onRelease = bh_reset_tapped,
         x = x,
         y = y
         } )
   
end -- create_reset proc
-----------------------------------------------------------------
function create_tap_me ()   -- button to handle a tap of a tableview row

   local x, y

   x = contW2
   y = contH2

   b_tap_me = widget.newButton ( {
         defaultFile = "b120_base.png",
         overFile = "b120_over.png",
         label = "Tap Me",
         emboss = true,
         fontSize = 30,
         onRelease = bh_button_tapped,
         x = x,
         y = y
         } )
   
   b_tap_me.isVisible = false

end -- create_tap_me proc
-----------------------------------------------------------------
function create_tv_list ()

   --------------------------------
   local function handle_render_a_row (event)   -- called to create text/graphics for each row of table, as scrolled into view
      local row, line_stroke_color
      local params, inx = 0
      local fsize_table_main 

      row = event.row
      params = row.params
      inx = params.inx

      line_stroke_color = { 0, 0, 0} -- black.   1,0,0.5 is pink 

      row.bg = display.newRect (0, 0, display.contentWidth, 60)   -- x, y, w, h
      row.bg.anchorX = 0
      row.bg.anchorY = 0
      row.bg.stroke = line_stroke_color
      row.bg.strokeWidth = 2   -- thickness of horizontal line separating items
      row.bg:setFillColor (1, 1, 1)   --   1,1,1 is white  background color (below text)
      row:insert (row.bg)
 
      fsize_table_main = 20
      row.line1_text = display.newText (tostring (inx) .. ": sample", 12, 0, "Helvetica", 30)
      row.line1_text.anchorX = 0
      row.line1_text.anchorY = 0.5
      row.line1_text:setFillColor (0)
      row.line1_text.y = 12   -- distance from horizontal line's top side (roughly) to start of text
                            -- in actuality, 0 seems to start it about 1/3 of a character *above* the line :(
      row.line1_text.x = 4    -- distance from left edge of rectangle
    
      fsize_table_main = 16
      row.line2_text = display.newText ("second row", 12, 0, "Helvetica", 24)
      row.line2_text.anchorX = 0
      row.line2_text.anchorY = 0.5
      row.line2_text:setFillColor (0.5)
      row.line2_text.y = 40
      row.line2_text.x = 20 
    
      row:insert (row.line1_text)
      row:insert (row.line2_text)
   
      return true

   end -- handle_render_a_row proc
   -----------------------------------
   local function handle_scroll_table (event)

      local row = event.row
      local phase = event.phase

      if phase == "began" then
         -- print ("began scroll")

      elseif phase == "moved" then
         -- print ("moved")

      elseif phase == "ended" then
         -- print ("ended")

      elseif phase == "swipeLeft" then
         print ("Swiped Left row: " .. inx)
         --row.reRender = true

      elseif phase == "swipeRight" then
         print ("Swiped Right row: " .. inx)

      else
         print ("phase is " .. tostring (phase))   -- can be nil sometimes!
             -- note: we see 'nil' and 'stopped' when scrolling to the end or start
      end

      return true

   end -- handle_scroll_table Proc
   -----------------------------------
   local function handle_touch_row (event)   -- when a row is tapped, this gets called

      local row = event.row
      local phase = event.phase     -- usually 'tap'
      local msecs, now

      inx = row.params.inx

      print ("handle_touch_row: ", inx)

          -- hide the table view, reveal a button...

      tv_list.isVisible = false
      b_tap_me.isVisible = true

      show_msg ("A row was tapped")

         -- kludge to avoid bounce

      now = system.getTimer ()
      msecs = now - avoid_bounce_time
      if msecs < 10 then
         print ("   ignore bounce *******************")
         avoid_bounce_time = 0

         show_msg ("OOPS: spurious handle_touch_row event!\n" ..
                   "Note: button should be invisible now, and\n" ..
                   "tableview should be visible.  This isn't\n" ..
                   "always the case (another bug)")

         return true
      end

        -- phases seen: tap, press, release, cancelled (not in order)

      if phase == "press" then
         print ("Pressed row: " .. inx)

      elseif phase == "release" or phase == "tap" then
         print ("Tapped and/or Released row")
 
           -- note: it is difficult to get a 'swipe' in the simulator!

      elseif phase == "swipeLeft" then
         print ("Swiped Left row")

      elseif phase == "swipeRight" then
         print ("Swiped Right row")
      end

   print ("  end handle touch row")

   return true   -- SS: should we do this?  I added it (not in demo example)

   end -- handle_touch_row Proc
   -----------------------------------

      -- create a table view, centered in middle of screen.
      -- The height should be about 1/2 the screen height 
      -- (leaving room above and below for other stuff).
      -- The width is about 2/3 of the screen width (ditto).
 
      -- Note: config file says screen is 640 x 960.
      -- That results in:
      --    Display content: width = 640; height = 960
      --    new table view: top = 240; width = 422; height = 480
      -- Observe that the tableview seems to have 10 pixels of nothing
      -- above the top row for some reason (i.e., if you ran the
      -- create_numbers_on_side function, you see that the table starts
      -- on screen at 250, not 240.  Weird.
 
       -- middle of tableview is middle of screen;
       -- and since we want it to take 1/2 the screen,
       -- then the top is 1/4 above the middle (and
       -- the bottom is 1/4 below the middle)

   top = math.floor (display.contentHeight * 0.25)   
 
   height = math.floor (display.contentHeight * 0.5)    -- half the screen height

   width = math.floor (display.contentWidth * 0.66)     -- 2/3 the screen width

   print ("Display content: width = " .. display.contentWidth .. "; height = " .. display.contentHeight)
   print ("new table view: top = " .. top .. "; width = " .. width .. "; height = " .. height)

   tv_list = widget.newTableView {
         top = top,
         width = width,
         height = height,
         x = contW2,         -- center of tableview is center of screen
         y = contH2,         -- center of tableview is center of screen
         onRowRender = handle_render_a_row,   -- invoked by 'insertRow' below
         onRowTouch = handle_touch_row,
         listener = handle_scroll_table,   -- gets 'began', 'moved', 'ended' phases
      }

   print ("tv x = " .. tv_list.x .. "; y = " .. tv_list.y)   -- will be 320, 480 on a 640 x 960 device

   for inx = 1, 20 do
      tv_list:insertRow {    -- calls handle_render_a_row for each row inserted
               params = {inx = inx},
               rowHeight = 60,
               }
   end      -- for 1..num_recs
  
end -- create_tv_list
-----------------------------------------------------------------
function reset_stuff ()

      -- set visibility of things to their starting state

   b_tap_me.isVisible = false
   t_help.isVisible = false
   t_msg.isVisible = true
   tv_list.isVisible = true

end -- reset_stuff proc
-----------------------------------------------------------------
function set_all_invisible ()

      -- set visibility of all things to false

   b_tap_me.isVisible = false
   t_help.isVisible = false
   t_msg.isVisible = false
   tv_list.isVisible = false

end -- set_all_invisible proc
-----------------------------------------------------------------
function show_msg (s)

   if t_msg then
      t_msg.text = s
   end

end -- show_msg proc
-----------------------------------------------------------------

t_title = display.newText ("TableView Bug Demo   " .. sversion, contW2, 40, "Helvetica", 30)
t_title:addEventListener ("touch", bh_title_text)

-- create_numbers_on_bottom ()   -- see comment at start (for debugging screen layout)
-- create_numbers_on_side ()     -- see comment at start (for debugging screen layout)

create_msg ()         -- the message text area, near bottom of screen
create_tap_me ()      -- the 'Tap Me' button, mid screen
create_tv_list ()     -- the table view, mid screen
create_reset ()       -- the 'Reset' button, lower left on screen
create_help ()        -- the 'Help' button, lower right on screen

--//
